


/*
 *  $Id: chrecv.c,v 1.25 1994/08/12 20:39:51 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: chrecv.c,v 1.25 1994/08/12 20:39:51 gropp Exp $";
#endif /* lint */

#include "mpid.h"
#ifndef MPI_ANY_TAG
#define MPI_ANY_SOURCE 	(-2)
#define MPI_ANY_TAG	(-1)
#endif

/* 
  This file contains the routines to handle receiving a message

  Because we don't know the length of messages (at least the long ones with
  tag MPID_PT2PT2_TAG(src) tags), we never post a receive.  Rather, we have
  a MPID_ch_check_incoming routine that looks for headers.  Note that
  messages sent from a source to destination with the MPID_PT2PT2_TAG(src)
  are ordered (we assume that the message-passing system preserves order).

  Possible Improvements:
  This current system does not "prepost" Irecv's, and so, on systems with
  aggressive delivery (like the Paragon), can suffer performance penalties.
  Obvious places for improvements are
     On blocking receives for long messages, post the irecv FIRST.
     If the message is actually short, cancel the receive.
         (potential problem - if the next message from the same source is
         long, it might match (incorrectly) with the posted irecv.  
         Possible fix: use sequence number for each processor pair, with
         the sequence numbers incremented for each message, short or long, 
         and place the sequence number into the tag field)

     For tags/sources/contexts in a specified range, post the irecv with
     at tag synthesized from all of the tag/source/context(/sequence number)
     May need to use cancel if the message is shorter than expected.
     This can be done with both blocking and non-blocking messages.

     Another approach is to generate "go-ahead" messages, to be handled in 
     chsend.c, perhaps by an interrupt-driven receive.  
     
 */


/***************************************************************************/
/* These routines enable the debugging output                              */
/***************************************************************************/
static int DebugFlag = 0;

void MPID_SetDebugFlag( f )
int f;
{
DebugFlag = f;
}
void MPID_SetRecvDebugFlag( f )
int f;
{
DebugFlag = f;
MPID_SetSyncDebugFlag( f );
}
/***************************************************************************/

/***************************************************************************/
/* These are used to keep track of the number and kinds of messages that   */
/* are received                                                            */
/***************************************************************************/
static int n_short       = 0,         /* short messages */
           n_long        = 0,         /* long messages */
           n_unexpected  = 0,         /* unexpected messages */
           n_syncack     = 0;         /* Syncronization acknowledgments */
static int DebugMsgFlag = 0;
int MPID_SetMsgDebugFlag( f )
int f;
{
DebugMsgFlag = f;
}
int MPID_GetMsgDebugFlag()
{
return DebugMsgFlag;
}
int MPID_PrintMsgDebug()
{
fprintf( stdout, "[%d] short = %d, long = %d, unexpected = %d, ack = %d\n",
	 __MYPROCID, n_short, n_long, n_unexpected, n_syncack );
}
/***************************************************************************/

/* Unpack a short message from a packet.
   If address is provided, use that for the data.  Should be
   mpid_recv_handle->start or an allocated location */
void MPID_EUI_extract_short( address, dmpi_recv_handle, len, inbuf )
void         *address, *inbuf;
int          len;
MPIR_RHANDLE *dmpi_recv_handle;
{
n_short++;
if (address) {
    memcpy( address, inbuf, len );
    }
else {
    DMPI_put_totallen( dmpi_recv_handle, len );
    DMPI_put_from_contig( dmpi_recv_handle, inbuf, len );
    }
if (DebugFlag) {
    MPID_EUI_Print_pkt_data( "R Got data from packet", address, len );
    }
}

/* Extract a long message from a packet
   If address is provided, use that for the data.  Should be
   mpid_recv_handle->start or an allocated location.

   inbuf is provided in case the chosen protocol places some data in the
   initial packet. 
 */
void MPID_EUI_extract_long( address, 
			   dmpi_recv_handle, len, inbuf, from )
void         *address, *inbuf;
int          len, from;
MPIR_RHANDLE *dmpi_recv_handle;
{
int  actual_len;
void *laddress;

n_long++;             
if (address) {
    {int __l=len;__EUIFROM=-1;__EUITYPE=MPID_PT2PT2_TAG(from);
mp_brecv((char *)address,&__l,&__EUIFROM,&__EUITYPE,&__EUILEN);};
    }
else {
    DMPI_put_totallen( dmpi_recv_handle, len );
    actual_len = len;
    DMPI_put_into_contig( dmpi_recv_handle, (void **)(&laddress), 
			  len, &actual_len );
    {int __l=actual_len;__EUIFROM=-1;__EUITYPE=MPID_PT2PT2_TAG(from);
mp_brecv(laddress,&__l,&__EUIFROM,&__EUITYPE,&__EUILEN);};
    }
if (DebugFlag) {
    printf( "[%d]R Got data from packet (%s:%d)\n", 
	    __MYPROCID, __FILE__, __LINE__ );
    printf( "[%d]R+ Receiving a long message from %d\n", __MYPROCID, from );
    fflush( stdout );
    }
}

#ifndef PI_NO_NRECV
/* Extract a long message using nonblocking operations
   nextract starts the operation, 
   wextract completes it.

   NOT YET USED!
 */
void MPID_EUI_nextract_long( address, 
			   dmpi_recv_handle, len, inbuf, from )
void         *address, *inbuf;
int          len, from;
MPIR_RHANDLE *dmpi_recv_handle;
{
int           actual_len;
void          *laddress;
int rc;

n_long++;             
if (address) {
    {int __l = len; __EUIFROM=-1;__EUITYPE=MPID_PT2PT2_TAG(from);
     mp_recv((char *)address,&__l,&__EUIFROM,&__EUITYPE,&(rc ));};
    }
else {
    DMPI_put_totallen( dmpi_recv_handle, len );
    actual_len = len;
    DMPI_put_into_contig( dmpi_recv_handle, (void **)(&laddress), 
			  len, &actual_len );
    {int __l = actual_len; __EUIFROM=-1;__EUITYPE=MPID_PT2PT2_TAG(from);
     mp_recv(laddress,&__l,&__EUIFROM,&__EUITYPE,&(rc ));};
    }
if (DebugFlag) {
    printf( "[%d]R Got data from packet (%s:%d)\n", 
	    __MYPROCID, __FILE__, __LINE__ );
    printf( "[%d]R+ Receiving a long message from %d\n", __MYPROCID, from );
    fflush( stdout );
    }
dmpi_recv_handle->dev_rhandle.rid = rc;
}
void MPID_EUI_wextract_long( dmpi_recv_handle )
MPIR_RHANDLE *dmpi_recv_handle;
{
mp_wait(&(dmpi_recv_handle->dev_rhandle.rid ),&__EUILEN);;
}
#endif

/*
   This code is called when a receive finds that the message has already 
   arrived and has been placed in the unexpected queue.  This code
   stores the information about the message (source, tag, length),
   copies the message into the receiver's buffer, and generates a
   acknowledgement if the message has mode SYNC.

   dmpi_recv_handle is the API's receive handle that is to receive the
   data.

   dmpi_unexpected is the handle of the data found in the unexpected queue.
 */
int MPID_EUI_Process_unexpected( dmpi_recv_handle, dmpi_unexpected )
MPIR_RHANDLE *dmpi_recv_handle, *dmpi_unexpected;
{
MPID_RHANDLE *mpid_recv_handle;
MPID_RHANDLE *mpid_recv_handle_unex;
int errno = MPI_SUCCESS;

n_unexpected++;

if (DebugFlag) {
    printf( "[%d]R Found message in unexpected queue (%s:%d)\n", 
	    __MYPROCID, __FILE__, __LINE__ );
    fflush( stdout );
    }
/* Copy relevant data to recv_handle */
mpid_recv_handle	   = &dmpi_recv_handle->dev_rhandle;
mpid_recv_handle_unex	   = &dmpi_unexpected->dev_rhandle;
dmpi_recv_handle->source   = dmpi_unexpected->source;
dmpi_recv_handle->tag	   = dmpi_unexpected->tag;
dmpi_recv_handle->totallen = mpid_recv_handle_unex->bytes_as_contig;
if (DebugFlag) {
    printf( "[%d]R Found message in temp area of %d bytes (%s:%d)...\n", 
	    __MYPROCID, mpid_recv_handle_unex->bytes_as_contig,
	    __FILE__, __LINE__ );
    fflush( stdout );
    }
/* Error test on length of message */
if (mpid_recv_handle->bytes_as_contig < dmpi_recv_handle->totallen) {
    mpid_recv_handle_unex->bytes_as_contig = mpid_recv_handle->bytes_as_contig;
    errno = MPI_ERR_TRUNCATE;
    (*MPID_ErrorHandler)( 1, "Truncated message"  );
    }
if (mpid_recv_handle_unex->bytes_as_contig > 0) {
    if (mpid_recv_handle->start) {
	if (DebugFlag) {
	    printf( "[%d]R About to copy to %x from %x (%s:%d)...\n", 
		    __MYPROCID,
		   mpid_recv_handle_unex->start, mpid_recv_handle_unex->temp,
		   __FILE__, __LINE__ );
	    fflush( stdout );
	    }
	memcpy( mpid_recv_handle->start, mpid_recv_handle_unex->temp,
	       mpid_recv_handle_unex->bytes_as_contig );
	}
    else {
	if (DebugFlag) {
	    printf( "[%d]R About to use DMPI to copy from %x (%s:%d)...\n", 
		   __MYPROCID, mpid_recv_handle_unex->temp,
		   __FILE__, __LINE__ );
	    fflush( stdout );
	    }
	DMPI_put_totallen( dmpi_recv_handle, 
			  mpid_recv_handle_unex->bytes_as_contig );
	DMPI_put_from_contig( dmpi_recv_handle, 
			     (void *)(mpid_recv_handle_unex->temp),
			     mpid_recv_handle_unex->bytes_as_contig );
	}
    }
if (DebugFlag) {
    printf( 
  "[%d]R Copied message out of temp area; send mode is %x (%s:%d)..\n", 
	    __MYPROCID, mpid_recv_handle_unex->mode, __FILE__, __LINE__ );
    fflush( stdout );
    }

if (mpid_recv_handle_unex->temp) {
    free(mpid_recv_handle_unex->temp );
    }
mpid_recv_handle_unex->temp = 0;      /* In case of a cancel */

/* Return the synchronization message */
if (MPIR_MODE_IS_SYNC(mpid_recv_handle_unex)) {
    if (DebugFlag) {
	printf( 
       "[%d]SYNC Returning sync for %x to %d for rcv of unxpcted (%s:%d)\n", 
	       __MYPROCID,
	        mpid_recv_handle_unex->mode, mpid_recv_handle_unex->from,
	        __FILE__, __LINE__ );
	fflush( stdout );
	}
    n_syncack++;
    MPID_SyncReturnAck( MPIR_MODE_SYNC_ID(mpid_recv_handle_unex),
		        mpid_recv_handle_unex->from );
    }

mpid_recv_handle->done = MPIR_YES;
DMPI_mark_recv_completed(dmpi_recv_handle);

/* Recover dmpi_unexpected */
DMPI_free_unexpected( dmpi_unexpected );
if (DebugFlag) {
    printf( "[%d]R Leaving 'process unexpected' (%s:%d)...\n", 
	    __MYPROCID, __FILE__, __LINE__ );
    fflush( stdout );
    }
return errno;
}


/*
   Post a receive.

   Since the Chameleon implementation lets the underlying message transport
   layer handle this, there isn't much to do.  We DO handle the case of
   posting a BLOCKING receive by doing BLOCKING waits on incoming messages.

   Otherwise, we simply try to handle any receives that are ready for
   processing.
 */
int MPID_EUI_post_recv( dmpi_recv_handle, is_available ) 
MPIR_RHANDLE *dmpi_recv_handle;
int          *is_available;
{
MPID_RHANDLE *mpid_recv_handle;
MPID_RHANDLE *mpid_recv_handle_unex;
MPIR_RHANDLE *dmpi_unexpected;
int          found;

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
*is_available    = 0;

if (DebugFlag) {
    printf( 
      "[%d]R starting recv for tag = %d, source = %d, ctx = %d (%s:%d)\n", 
	    __MYPROCID, dmpi_recv_handle->tag, dmpi_recv_handle->source,
	    dmpi_recv_handle->contextid, __FILE__, __LINE__ );
    fflush( stdout );
    }
/* At this time, we check to see if the message has already been received.
   (this is a macro that checks first to see if the queue is empty) */
DMPI_search_unexpected_queue( dmpi_recv_handle->source, 
		   dmpi_recv_handle->tag, dmpi_recv_handle->contextid, 
		   &found, 1, &dmpi_unexpected );
if (found) {
    if (DebugFlag) {
	printf( "[%d]R found in unexpected queue (%s:%d)\n", 
	        __MYPROCID, __FILE__, __LINE__ );
	fflush( stdout );
	}
    MPID_EUI_Process_unexpected( dmpi_recv_handle, dmpi_unexpected );
    *is_available = 1;
    return MPI_SUCCESS;
    }

/* Add to the posted receive queue */
MPIR_enqueue( &MPIR_posted_recvs, dmpi_recv_handle, MPIR_QRHANDLE );

/* If we got here, the message is not yet available */
if (!mpid_recv_handle->is_non_blocking) {
    if (DebugFlag) {
	printf( "[%d]R Blocking recv; starting wait loop (%s:%d)\n", 
	        __MYPROCID, __FILE__, __LINE__ );
	fflush( stdout );
	}
    /* Continue to process incoming messages until the one we want 
       arrives */
    while (!dmpi_recv_handle->completed) {
	(void)MPID_EUI_check_incoming( MPID_BLOCKING );
	}
    }

else {
    if (DebugFlag) {
	printf( 
       "[%d]R About to do a non-blocking check of incoming messages (%s:%d)\n",
	        __MYPROCID, __FILE__, __LINE__ );
	fflush( stdout );
	}
    /* Process all pending messages until there are none left */
    while (MPID_EUI_check_incoming( MPID_NOTBLOCKING ) != -1) ;
    }

if (DebugFlag) {
    printf( "[%d]R Exiting post receive (%s:%d)\n", 
	    __MYPROCID, __FILE__, __LINE__ );
    fflush( stdout );
    }

return MPI_SUCCESS;
}

/*
   Copy the body of a message into the destination buffer for a posted
   receive 
 */
int MPID_EUI_Copy_body( dmpi_recv_handle, pkt, from )
MPIR_RHANDLE *dmpi_recv_handle;
MPID_PACKET  *pkt;
int          from;
{
MPID_RHANDLE *mpid_recv_handle;
int          msglen;

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
msglen = MPID_MSG_LEN(pkt);

dmpi_recv_handle->totallen = msglen;
if (msglen <= MPID_PACKET_SIZE) 
    MPID_EUI_extract_short( mpid_recv_handle->start, 
			  dmpi_recv_handle, msglen, pkt->buffer );
else {
    MPID_EUI_extract_long( mpid_recv_handle->start, 
			 dmpi_recv_handle, msglen, pkt->buffer, from );
    }
if (MPID_MODE_IS_SYNC(pkt)) {
    if (DebugFlag) {
	printf( "[%d]SYNC Returning sync to %d with mode ", __MYPROCID,
	       from );
	MPID_Print_mode( stdout, pkt );
	fprintf( stdout, "(%s:%d)\n", __FILE__, __LINE__ );
	fflush( stdout );
	}
    n_syncack++;
    MPID_SyncReturnAck( MPID_SYNC_GET_ID( pkt ), from );
    }
DMPI_mark_recv_completed(dmpi_recv_handle);
mpid_recv_handle->done = MPIR_YES;

return MPI_SUCCESS;
}

/*
   Copy the body of a message into the destination buffer for an
   unexpected message.
 */
int MPID_EUI_Copy_body_unex( dmpi_recv_handle, pkt, from )
MPIR_RHANDLE *dmpi_recv_handle;
MPID_PACKET  *pkt;
int          from;
{
MPID_RHANDLE *mpid_recv_handle;
char *address;
int  msglen;

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
msglen = MPID_MSG_LEN(pkt);

/* We buffer the message on the receiver.   There are, of course, other
   strategies */
/* We want to avoid allocating storage of size zero */
if (msglen > 0) {
    address = (char *)malloc(msglen ); 
    if (!address) {
	(*MPID_ErrorHandler)( 1, 
			 "No more memory for storing unexpected messages"  );
	return MPI_ERR_EXHAUSTED;
	}
    if (msglen <= MPID_PACKET_SIZE) 
	MPID_EUI_extract_short( address, (MPID_RHANDLE *)0, 
			      msglen, pkt->buffer );
    else 
	MPID_EUI_extract_long( address, (MPID_RHANDLE *)0, 
			     msglen, pkt->buffer, from );
    }
else 
    address = (void *)0;
#ifdef FOO
if (DebugFlag) {
    printf( "[%d]SYNC Returning sync to %d with mode ", __MYPROCID,
	   from );
    MPID_Print_mode( stdout, pkt );
    fprintf( stdout, "(%s:%d)\n", __FILE__, __LINE__ );
    fflush( stdout );
    }
#endif
mpid_recv_handle->temp            = address;
mpid_recv_handle->bytes_as_contig = msglen;

/* We also need to remember that the mode of this message in case
   it was a Synchronous message or encoded with XDR */
mpid_recv_handle->mode = MPID_MODE_GET_SYNC_WITH_ID(pkt);
mpid_recv_handle->from = from;
#ifdef MPID_HAS_HETERO
if (MPID_MODE_HAS_XDR(pkt)) 
    dmpi_recv_handle->msgrep = MPIR_MSGREP_XDR;
else 
    dmpi_recv_handle->msgrep = MPIR_MSGREP_RECEIVER;
#endif

#ifdef DEBUG_READY
if (MPID_MODE_IS_READY(pkt)) {
    (*MPID_ErrorHandler)( 1, 
			 "Received ready message without matching receive"  );
    return MPI_ERR_NOMATCH;
    }
#endif
return MPI_SUCCESS;
}

/* Check for incoming messages.
    Input Parameter:
.   is_blocking - true if this routine should block until a message is
    available

    Returns -1 if nonblocking and no messages pending
 */    
int MPID_EUI_check_incoming( is_blocking )
MPID_BLOCKING_TYPE is_blocking;
{
MPID_PACKET  pkt;	
int          from;
MPIR_RHANDLE *dmpi_recv_handle;
int          actual_len;
MPID_RHANDLE *mpid_recv_handle;
int          is_posted;

/* If nonblocking and no headers available, exit */
#ifndef pvm3
if (is_blocking == MPID_NOTBLOCKING) {
    if (!(__EUIFROM=-1,__EUITYPE=MPID_PT2PT_TAG ,
mp_probe(&__EUIFROM,&__EUITYPE,&__EUILEN),(__EUILEN>=0))) return -1;
    }
{int __l=sizeof(MPID_PACKET);__EUIFROM=-1;__EUITYPE=MPID_PT2PT_TAG;
mp_brecv(&pkt,&__l,&__EUIFROM,&__EUITYPE,&__EUILEN);};
#else
/* pvm3.0 doesn't have a real probe, but what they do have meets the 
   semantics that we need here, though it is somewhat painful... 
   All this to save  the user a single routine call in the case where
   a probe is immediately followed by a recv.  Heaven help you if you
   use the probe to decide to call some other code to process the 
   message... 
*/
{
int bufid, bytes, msgtype; 
if (is_blocking == MPID_NOTBLOCKING) {
    if ((bufid = pvm_nrecv( -1, MPID_PT2PT_TAG )) <= 0) return -1;
    /* If we found a message, we now have to receive it */
    pvm_bufinfo( bufid, &bytes, &msgtype, &__PVMFROMTID );
    pvm_upkint( (int *)&pkt, bytes / sizeof(int), 1 );
    __PVMFROM = -1;
    }
else {
    /* For the blocking case, we can use the existing code ... */
    {int __l=sizeof(MPID_PACKET);__EUIFROM=-1;__EUITYPE=MPID_PT2PT_TAG;
mp_brecv(&pkt,&__l,&__EUIFROM,&__EUITYPE,&__EUILEN);};
    }
}
#endif

from = __EUIFROM;
/* We should check the size here for internal errors .... */

if (DebugFlag) {
    printf( 
   "[%d]R received message (%s:%d)\n", __MYPROCID, __FILE__, __LINE__ );
    MPID_Print_packet( stdout, &pkt );
    }

if (MPID_MODE_IS_SYNCACK(&pkt)) {
    MPID_SyncAck( &pkt, from );
    return MPI_SUCCESS;    
    }

if (DebugFlag) {
    printf( 
   "[%d]R received message for tag = %d, source = %d, ctx = %d, mode = ", 
	    __MYPROCID, pkt.tag, from, pkt.context_id );
    MPID_Print_mode( stdout, &pkt );
    fprintf( stdout, "(%s:%d)\n", __FILE__, __LINE__ );
    fflush( stdout );
    }
/* Is the message expected or not? 
   This routine RETURNS a dmpi_recv_handle, creating one if the message 
   is unexpected (is_posted == 0) */
DMPI_msg_arrived( pkt.lrank, pkt.tag, pkt.context_id, 
                  &dmpi_recv_handle, &is_posted );
mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
if (DebugFlag) {
    printf( "[%d]R msg was %s (%s:%d)\n", __MYPROCID, 
	    is_posted ? "posted" : "unexpected", __FILE__, __LINE__ );
    }
if (is_posted) {
    MPID_EUI_Copy_body( dmpi_recv_handle, &pkt, from );
    }
else {
    MPID_EUI_Copy_body_unex( dmpi_recv_handle, &pkt, from );
    }
return MPI_SUCCESS;    
}


/*
    This routine completes a particular receive.  It does this by processing
    incoming messages until the indicated message is received.

    For fairness, we may want a version with an array of handles.
 */
int MPID_EUI_complete_recv( dmpi_recv_handle ) 
MPIR_RHANDLE *dmpi_recv_handle;
{
MPID_RHANDLE *mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;

while (!mpid_recv_handle->done) {
    (void)MPID_EUI_check_incoming( MPID_BLOCKING );
    }
return MPI_SUCCESS;
}


/*
   Special case code for blocking receive.  The "common" case is handled with
   straight-through code; uncommon cases call routines.
 */
int MPID_EUI_blocking_recv( dmpi_recv_handle ) 
MPIR_RHANDLE *dmpi_recv_handle;
{
MPID_RHANDLE *mpid_recv_handle, *mpid_recv_handle_unex;
MPIR_RHANDLE *dmpi_unexpected, *dmpi_save_recv_handle;
MPID_PACKET  pkt;	
int          found, from, actual_len, is_posted, tag, source, context_id;

mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;

if (DebugFlag) {
    printf( 
 "[%d]R starting blocking recv for tag = %d, source = %d, ctx = %d (%s:%d)\n", 
	    __MYPROCID, dmpi_recv_handle->tag, dmpi_recv_handle->source,
	    dmpi_recv_handle->contextid, __FILE__, __LINE__ );
    fflush( stdout );
    }
/* At this time, we check to see if the message has already been received */
DMPI_search_unexpected_queue( dmpi_recv_handle->source, 
		   dmpi_recv_handle->tag, dmpi_recv_handle->contextid, 
		   &found, 1, &dmpi_unexpected );
if (found) {
    MPID_EUI_Process_unexpected( dmpi_recv_handle, dmpi_unexpected );
    return MPI_SUCCESS;
    }

dmpi_save_recv_handle = dmpi_recv_handle;
/* If we got here, the message is not yet available */
if (DebugFlag) {
    printf( "[%d]R Blocking recv; starting wait loop (%s:%d)\n", 
	    __MYPROCID, __FILE__, __LINE__ );
    fflush( stdout );
    }
tag	   = dmpi_recv_handle->tag;
context_id = dmpi_recv_handle->contextid;
source	   = dmpi_recv_handle->source;
while (!dmpi_save_recv_handle->completed) {
    {int __l=sizeof(MPID_PACKET);__EUIFROM=-1;__EUITYPE=MPID_PT2PT_TAG;
mp_brecv(&pkt,&__l,&__EUIFROM,&__EUITYPE,&__EUILEN);};
    from = __EUIFROM;
    /* We should check the size here for internal errors .... */

    if (MPID_MODE_IS_SYNCACK(&pkt)) {
	MPID_SyncAck( &pkt, from );
	continue;
	}
    if (DebugFlag) {
	printf( 
       "[%d]R received message for tag = %d, source = %d, ctx = %d (%s:%d)\n", 
	       __MYPROCID, pkt.tag, from, pkt.context_id, __FILE__, __LINE__ );
	fflush( stdout );
	}
    if (pkt.context_id == context_id && 
	(pkt.tag   == tag    || tag == MPI_ANY_TAG) &&
	(pkt.lrank == source || source == MPI_ANY_SOURCE)) {
	/* Found the message that I'm waiting for (it was never queued) */
	is_posted = 1;
	dmpi_recv_handle	 = dmpi_save_recv_handle;
	dmpi_recv_handle->tag	 = pkt.tag;
	dmpi_recv_handle->source = pkt.lrank;
	}
    else {
	/* Unexpected message */
	DMPI_msg_arrived( pkt.lrank, pkt.tag, pkt.context_id, 
			  &dmpi_recv_handle, &is_posted );
	}
    mpid_recv_handle = &dmpi_recv_handle->dev_rhandle;
#ifdef MPID_HAS_HETERO
    /* Look for XDR bit */
    if (MPID_MODE_HAS_XDR(&pkt)) 
	dmpi_recv_handle->msgrep = MPIR_MSGREP_XDR;
    else
	dmpi_recv_handle->msgrep = MPIR_MSGREP_RECEIVER;
#endif
    if (DebugFlag) {
	printf( "[%d]R msg was %s (%s:%d)\n", __MYPROCID, 
	        is_posted ? "posted" : "unexpected", __FILE__, __LINE__ );
	}
    if (is_posted) {
	MPID_EUI_Copy_body( dmpi_recv_handle, &pkt, from );
	}
    else {
	MPID_EUI_Copy_body_unex( dmpi_recv_handle, &pkt, from );
	}
    }
}


