/* 
   This file defines the packet/message format that WILL be used in
   the next generation ADI.  
 */

#ifndef MPID_PKT_DEF
#define MPID_PKT_DEF
/* Here are all of the packet types.  We use the first bit to indicate
   short or long; by using small integers, we can use a single 
   select statement to jump to the correct code (here's hoping the 
   compiler generates good code for that!).
   In the cases where the packed is a control packet (neither long nor short),
   the first bit does NOT mean long/short.
 */
typedef enum { MPID_PKT_SHORT=0, MPID_PKT_LONG=1, MPID_PKT_SHORT_SYNC=2,
	       MPID_PKT_LONG_SYNC=3, MPID_PKT_SHORT_READY=4, 
               MPID_PKT_LONG_READY=5, 
               MPID_PKT_SYNC_ACK=6, MPID_PKT_REQUEST_SEND=7,
               MPID_PKT_OK_TO_SEND=8, 
	       MPID_PKT_READY_ERROR=9, 
	       MPID_PKT_COMPLETE_SEND=10, MPID_PKT_COMPLETE_RECV=11 } 
    MPID_Pkt_t;

/* Comments on packets */
/* The design here is fairly general, allowing both eager and reluctant 
   (rendevous) protocols for each of the MPI modes.  Some simplification can
   be achieved if, for example, the MPI Synchronous mode is only implemented
   with the reluctant protocol.

   Eager:
   In the eager protocol, the message data is sent without first asking the
   destination node (for example, to check on buffer space).  In order to 
   achieve better use of the underlying message-passing systems, there are
   two major categories: long and short.  In a short message, the message
   data and message envelope (context, tag, lrank) are sent together.
   In a long message, the data is sent separately from the envelope (avoiding
   any data copies).  If the MPI mode is Synchronous or Ready, then 
   some additional information is included.
   In the synchronous case, it is necessary for the receiver to tell the
   sender when the receive has started (but see Rendevous below); this
   is done with the MPID_PKT_SYNC_ACK.
   In the ready case, if the matching receive is NOT available, then an
   error message can be sent back to the sender with MPID_PKT_READY_ERROR.

   In the long message case, the SENDER may use monblocking send operation.
   On some systems, it may be necessary to ask the sender to execute an 
   MPI_Wait-like operation on the nonblocking send in order to get a 
   matching receive to complete.  On these systems, the macro
   MPID_WSEND_BEFORE_WRECV is defined, and the MPI_Wait on the
   send is requested with the MPID_PKT_COMPLETE_SEND packet.  (The systems
   that require this tend to use rendevous or reluctant protocols for long 
   messages).

   Rendevous:
   In the Rendevous protocol, a LONG send is not begun until the receiver
   oks it.  The request is made with MPID_PKT_REQUEST_SEND, and 
   is acknowledged with MPID_PKT_OK_TO_SEND.  Note that a message-passing
   implementation may make use of a "ready-receiver" send once the 
   OK_TO_SEND has been received.  Also note that the rendevous protocol
   handles standard and synchronous sends identically; no separate SYNC_ACK
   packet is required.  

   Undecided: should there be a REQUEST_TO_SEND_READY for ready-send mode?   

   Depending on the system, it may or may not be advantageous to compress
   the size of the data in the packet.  If the macro MPID_PKT_COMPRESSED
   is defined, then the envelope data is encoded into 2 ints.
 */
   
/* long and short forms of standard and ready are the same */
#ifdef MPID_PKT_COMPRESSED
#define MPID_PKT_MODE  unsigned mode:4;
#define MPID_PKT_BASIC \
    unsigned mode:4;             /* Contains MPID_Pkt_t */             \
    unsigned context_id:16;      /* Context_id */                      \
    unsigned lrank:12;           /* Local rank in sending context */   \
    int      tag;                /* tag is full sizeof(int) */
#else
#define MPID_PKT_MODE  int mode;
#define MPID_PKT_BASIC \
    MPID_Pkt_t  mode;       \
    int         context_id; \
    int         lrank;      \
    int         tag;  
#endif

#define MPID_PKT_DATA_SIZE 1024

/* 
   One unanswered question is whether it is better to send the length of
   a short message in the short packet types, or to compute it from the
   message-length provided by the underlying message-passing system.
   Currently, I'm planning to compute it.  Note that for short messages, I 
   only need another 2 bytes to hold the length (1 byte if I restrict
   short messages to 255 bytes).  The tradeoff here is addition computation
   at sender and receiver versus reduced data-load on the connection between
   sender and receiver.
 */
typedef struct { 
    MPID_PKT_BASIC
    char     buffer[MPID_PKT_DATA_SIZE];
    } MPID_PKT_SHORT_T;
typedef struct {
    MPID_PKT_BASIC
    int      len;                /* Number of bytes in message */
    } MPID_PKT_LONG_T;

typedef struct {
    MPID_PKT_BASIC
    int      sync_id;
    char     buffer[MPID_PKT_DATA_SIZE];
    } MPID_PKT_SHORT_SYNC_T;
typedef struct {
    MPID_PKT_BASIC
    int      len;
    int      sync_id;
    } MPID_PKT_LONG_SYNC_T;

typedef struct {
    MPID_PKT_MODE
    int      sync_id;
    } MPID_PKT_SYNC_ACK_T;

typedef struct {
    MPID_PKT_MODE
    int      send_id;
    } MPID_PKT_COMPLETE_SEND_T;
typedef struct {
    MPID_PKT_MODE
    int      recv_id;
    } MPID_PKT_COMPLETE_RECV_T;
typedef struct {
    MPID_PKT_BASIC
    int len;
    int send_id;
    } MPID_PKT_REQUEST_SEND_T;
typedef struct {
    MPID_PKT_MODE
    int send_id;
    int use_tag;
    } MPID_PKT_OK_TO_SEND_T;
typedef struct {
    MPID_PKT_BASIC
    int len;
    } MPID_PKT_READY_ERROR;

typedef union {
    MPID_PKT_SHORT_T         short_pkt;
    MPID_PKT_LONG_T          long_pkt;
    MPID_PKT_SHORT_SYNC_T    short_sync_pkt;
    MPID_PKT_LONG_SYNC_T     long_sync_pkt;
    MPID_PKT_SHORT_READY_T   short_ready_pkt;
    MPID_PKT_LONG_READY_T    long_ready_pkt;
    MPID_PKT_SYNC_ACK_T      sync_ack_pkt;
    MPID_PKT_COMPLETE_SEND_T send_pkt;
    MPID_PKT_COMPLETE_RECV_T recv_pkt;
    MPID_PKT_REQUEST_SEND_T  request_pkt;
    MPID_PKT_OK_TO_SEND_T    sendok_pkt;
    MPID_PKT_READY_ERROR_T   error_pkt;
    } MPID_PKT_T;


/* Managing packets

   In a perfect world, there would always be a place for an incoming packet
   to be received.  In systems that work at the hardware level, this is
   often managed by having a separate pool for each possible source, and
   having each pair of processors keep track of how much space is being used.
   In the implementations of the ADI on top of existing message-passing 
   systems, we usually allow the underlying message-passing system to 
   manage flow control.  

   The message-passing equivalent of having an available buffer is to pre-post
   a non-blocking receive into which an incoming message can be placed.  The
   pros of this are that unnecessary data movement (from internal to ADI's 
   buffers) can be avoided, and that systems with interrupt-driven 
   receives (e.g., Intel, TMC, IBM) can repsond on an interrupt basis
   to incoming packets.

   The con-side to this is that doing an Irecv/Wait pair can be more
   expensive than a (blocking) Recv, and that interrupts can be expensive.
   Since we intend to do both a native shared-memory and active-message 
   version, and since there probably isn't a correct answer to question of
   which approach is best, we provide for both based on whether the macro
   MPID_PKT_PRE_POST is defined.
   
   The operations are
   MPID_PKT_ALLOC() - Allocate the packets.  Either a single packet is
   used (non-pre-posted) or multiple packets (possibly two, to use double  
   buffering, for starters).

   MPID_PKT_INIT() - Initialize the packets .  Called during init portion.

   MPID_PKT_FREE() - Fress the allocated packets.  Note: if the packets
   are allocated on the calling-routines stack, this does nothing.
   
   MPID_PKT_CHECK() - Basically a check to see if a packet is available
 
   MPID_PKT_WAIT() - Waits for a packet to be available

   MPID_PKT_POST() - Post a non-blocking receive for the next packet.
   May be a nop in the case that packets are held by the underlying
   message-passing system

   MPID_PKT_POST_AND_WAIT() - Post and wait for a packet to be 
   available.  A blocking receive in both cases

 */
#ifdef MPID_PKT_PRE_POST
/* Single buffer for now. Note that this alloc must EITHER be in the
   same routine as all of the calls OR in the same file .
   An alternative is to make these POINTERS to global values
   declared elsewhere 
 */
#define MPID_PKT_ALLOC() \
 static MPID_Pkt_t     pkt; \
 static ASYNCRecvId_t  pktid;
#define MPID_PKT_INIT() MPID_PKT_POST()
#define MPID_PKT_CHECK()  PInstatus(&pktid)
#define MPID_PKT_WAIT() PIwrecv(0,0,0,0,&pktid)
#define MPID_PKT_POST() \
    PInrecv(MPID_PT2PT_TAG, &pkt, sizeof(MPID_Pkt_t), MSG_OTHER, &pktid)
#define MPID_PKT_POST_AND_WAIT() \
    PIbrecv(MPID_PT2PT_TAG, &pkt, sizeof(MPID_Pkt_t), MSG_OTHER )
#define MPID_PKT_FREE()
#else
#define MPID_PKT_ALLOC() MPID_Pkt_t pkt;
#define MPID_PKT_INIT()
#define MPID_PKT_CHECK()  PInprobe(MPID_PT2PT_TAG)
#define MPID_PKT_WAIT() MPID_PKT_POST_AND_WAIT()
#define MPID_PKT_POST() 
#define MPID_PKT_POST_AND_WAIT() \
    PIbrecv(MPID_PT2PT_TAG, &pkt, sizeof(MPID_Pkt_t), MSG_OTHER )
#define MPID_PKT_FREE()
#endif

#endif
